/* risc.cpp: RISC Compiler (c) 2015 CECS (TS, GL, RD) */

#include <iostream>

#include "risc.h"

#define RISC_VERBOSE            1
#define RISC_VERY_VERBOSE       2
#define RISC_VERY_VERY_VERBOSE  3
#define RISC_PEDANTIC           4

int main(int argc, char* argv[])
{
  std::map<SgNode*, int> FuncCallExptoID;
  risc::inst::ParseStatus status;

  status = risc::inst::ParseCommandlineOptions(argc, argv);

  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC Compiler starting..."
          << std::endl;
  }
  if (status.VerbosityLevel >= RISC_VERBOSE)
  { std::cout << "RISC: parsing, building ROSE IR..."
          << std::endl;
  }
  risc::Design d(argc,argv);

  if (status.VerbosityLevel >= RISC_VERBOSE)
  { std::cout << "RISC: analyzing, building RISC SG..."
          << std::endl;
  }
  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      d.systemc_definitions_->sc_wait_funcs_);
  risc::sg::SegmentGraph sg(function_call_boundary);

  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC: listing model hierarchy..."
          << std::endl;
  }
  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {
    if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
    { std::cout << "  Module: " <<  (*iter_module)->get_name() << std::endl;
    }
    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {
      if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
      { std::cout << "    " << (*iter_thread)->get_name() << std::endl;
      }
      sg.add_function(*iter_thread);
/************** Added by Guantao Liu 08/28/2015 **************/
      FuncCallExptoID[(*iter_thread)->starting_node_] =
                (*iter_thread)->starting_id_;
/**************** End  Guantao Liu 08/28/2015 ****************/
      }
    }
  sg.clean_graph();

  if (status.VerbosityLevel >= RISC_VERBOSE)
  { std::cout << "RISC: analyzing, building conflict tables..."
          << std::endl;
  }
  // Time advance tables
  risc::sg::TimeAdvancedTable time_advanced_table(sg);
  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC: generating \"current_time_advance_table.html\"..."
          << std::endl;
    time_advanced_table.print_current_time_advanced_table(
                          "current_time_advance_table.html");
    std::cout << "RISC: generating \"next_time_advance_table.html\"..."
          << std::endl;
    time_advanced_table.print_next_time_advanced_table(
                          "next_time_advance_table.html");
  }
  // Event conflict table
  risc::sg::EventConflictTable
    event_conflict_table(sg, d.path_instance_mapper_, true);
  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC: generating \"event_conflict_table.html\"..."
          << std::endl;
    event_conflict_table.print_conflict_table("event_conflict_table.html");
    std::cout << "RISC: generating \"conflict_lookup_table.html\"..."
          << std::endl;
    event_conflict_table.print_lookup_table("conflict_lookup_table.html");
  }
  // Data conflict table
  risc::sg::DataConflictTable
    data_conflict_table(sg, d.path_instance_mapper_, true);
  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC: generating \"data_conflict_table.html\"..."
          << std::endl;
    data_conflict_table.print_conflict_table("data_conflict_table.html");
  }
  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC: generating \"segment_graph.dot\"..."
          << std::endl;
    sg.print_graph("segment_graph.dot");
  }

/************** Added by Guantao Liu 08/24/2015 **************/
  if (status.VerbosityLevel >= RISC_VERBOSE)
  { std::cout << "RISC: instrumenting, generating parallel model..."
          << std::endl;
  }
  std::pair<risc::sg::VertexIterator, risc::sg::VertexIterator> vertices_range = vertices(sg.graph_);
  for (risc::sg::VertexIterator iter = vertices_range.first; iter != vertices_range.second; iter++)
  {
    const SgExprStatement* waitexprstmt = isSgExprStatement(sg.graph_[*iter].boundary_node_);
    if (waitexprstmt != NULL)
    {
      SgFunctionCallExp* funccallexp = isSgFunctionCallExp(waitexprstmt->get_expression());
      if (funccallexp != NULL)
        FuncCallExptoID[funccallexp] = sg.graph_[*iter].id_;
    }
  }
  risc::inst::instrumentor(d, &data_conflict_table, &event_conflict_table, &time_advanced_table, FuncCallExptoID,
        status.VerbosityLevel /* .... FIXME .... */);
/**************** End  Guantao Liu 08/24/2015 ****************/

  if (status.VerbosityLevel >= RISC_VERY_VERBOSE)
  { std::cout << "RISC Compiler done."
          << std::endl;
  }
  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
