#ifndef MODULE_H_INCLUDED_
#define MODULE_H_INCLUDED_

#include<iostream>
#include<string>
#include<vector>

#include "module_instance.h"
#include "variable.h"

#include "function.h"
#include "thread.h"
#include "cthread.h"
#include "method.h"

#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "primitive_channel.h"
#include "primitive_channel_instance.h"


#include "class.h"


namespace risc {

class Module;
typedef std::vector<Module*> ModuleVector;
typedef std::vector<Module*>::iterator ModuleVectorIter;
typedef std::vector<Module*>::const_iterator ModuleVectorConstIter;

/**
 *  \class Module
 *  \brief This class represents a definition of a module
 *
 *  A module is defined by the assembling of modules, channels, ports,
 *  processes, and others.
 */
class Module: public Class {

public:

  /**
   * Default constructor
   *
   * \brief Creates an empty module definition. All the lists of instances
   *        and sub definitions are empty.
   */
  explicit Module(SgClassDefinition *ast_node);

  ~Module();


  ///////////////////////// initialization funcitons /////////////////////////

  /**
   * \brief  This function searches for all functions, SC_THREADS, SC_CTHREADS,
   *         and SC_METHODS  in the module definition. The corresponding
   *         variable vectors will be used.
   */
  virtual void initialize_functions();

  /////////////////////////////  helper functions ////////////////////////////

  /**
   * \brief  Prints the complete module design into the terminal
   */
  void print_module_definition();

  /**
   * \brief  Prints the hierarchy. Roots is this definiton
   */
  void print_hierarchy(unsigned indent_level = 0);

  ////////////////////  Vectors module definition elements ///////////////////

  /**
   * \brief Vector of module definitions which are defined in this module.
   */
  ModuleVector sub_modules_defs_;

  /**
   * \brief Vector of primitive channel definitions. Technically, this
   *        represents a nested C++ class definition. In a good design this
   *        vector should be empty.
   */
  PrimitiveChannelVector sub_prim_channel_defs_;

  /**
   * \brief Vector of hierarchical channel definitions. Technically, this
   *        represents a nested C++ class definition. In a good design this
   *        vector should be empty.
   */
  HierarchicalChannelVector sub_hier_channel_defs_;

  /**
   * \brief Vector of threads which are defined in this module
   */
  ThreadVector threads_;

  /**
   * \brief Vector of cthreads which are defined in this module
   */
  CThreadVector cthreads_;

  /**
   * \brief Vector of methods which are defined in this module
   */
  MethodVector methods_;

  /**
   * \brief Vector of reset signals which are registered in this module
   */
  ResetSignalVector reset_signals_;

  /**
   * \brief Vector of async reset signals which are registered in this module
   */
  ResetSignalVector async_reset_signals_;

  /**
   * \brief This variable counts how many instances of this given type exsist
   *        in the desgin
   */
  int instance_counter_;

private:

  /**
   * \brief Disabled copy constructor
   */
  Module(const Module &md);
};

} // end of namespace risc

#endif /* MODULE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
