#include "path_instance_mapper.h"
#include "design.h"

risc::PathInstanceMapper::PathInstanceMapper():
  max_instances_(0)
{
}

void
risc::PathInstanceMapper::determine_lookup_table(InstanceTree root)
{
  risc::ModuleInstance *module_instance
    = dynamic_cast<risc::ModuleInstance*>(root.get_object());

  // So, we are dealing with a module instance
  if(module_instance) {

    Module *module = module_instance->definition_;

    // We are iterating over all declarations in the class/module.
    // We cannot use the internal infrastructure because we need to do it in
    // the depth first search order.
    for(SgDeclarationStatementPtrList::iterator
        iter  = module->get_ast_node()->get_members().begin();
        iter != module->get_ast_node()->get_members().end();
        iter++) {

      // we have to consider all variable declarations
      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      if(var_decl) {

        // here we are looking up if the variable declaration is a module
        ModuleInstance *module
          = module_instance->definition_->find_module(
                                              var_decl->get_definition());

        if(module) {

          // we are going recursively through the design
          risc::InstanceTree::Element sub_path;

          sub_path.first = module_instance->definition_;
          sub_path.second = module;

          root.push_back(sub_path);
          determine_lookup_table(root);
          root.pop_back();
          continue;
        }

        // Same for hierarchical channels
        HierarchicalChannelInstance *hier_channel_instance
          = module_instance->definition_->find_hier_channel(
                                              var_decl->get_definition());

        if(hier_channel_instance) {
          risc::InstanceTree::Element sub_path;

          sub_path.first = hier_channel_instance->definition_;
          sub_path.second = hier_channel_instance;

          root.push_back(sub_path);
          determine_lookup_table(root);
          root.pop_back();
        }
      }
    }
  }

  if(module_instance) {

    path_to_id_mapping_.push_back(
        PathWithID(root, module_instance->definition_->instance_counter_));

    module_instance->definition_->instance_counter_++;

    if(max_instances_ < module_instance->definition_->instance_counter_) {
      max_instances_ = module_instance->definition_->instance_counter_;
    }
    return;
  }


  // The following is undocumented because it is the same like above just like
  // for hierarchical channels
  risc::HierarchicalChannelInstance *hier_channel_instance
    = dynamic_cast<risc::HierarchicalChannelInstance*>(root.get_object());

  if(hier_channel_instance) {

    HierarchicalChannel *hier_channel = hier_channel_instance->definition_;

    for(SgDeclarationStatementPtrList::iterator
        iter  = hier_channel->get_ast_node()->get_members().begin();
        iter != hier_channel->get_ast_node()->get_members().end();
        iter++) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      if(var_decl) {

        ModuleInstance *module
          = hier_channel_instance->definition_->find_module(
                                              var_decl->get_definition());

        if(module) {

          risc::InstanceTree::Element sub_path;

          sub_path.first = module_instance->definition_;
          sub_path.second = module;

          root.push_back(sub_path);
          determine_lookup_table(root);
          root.pop_back();
        }


        HierarchicalChannelInstance *instance
          = hier_channel_instance->definition_->find_hier_channel(
                                                    var_decl->get_definition());

        if(instance) {
          risc::InstanceTree::Element sub_path;

          sub_path.first = hier_channel_instance->definition_;
          sub_path.second = instance;

          root.push_back(sub_path);
          determine_lookup_table(root);
          root.pop_back();
        }
      }
    }
  }

  if(hier_channel_instance) {

    path_to_id_mapping_.push_back(
        PathWithID(root, hier_channel_instance->definition_->instance_counter_));

    hier_channel_instance->definition_->instance_counter_++;

    if(max_instances_ < hier_channel_instance->definition_->instance_counter_) {
      max_instances_ = hier_channel_instance->definition_->instance_counter_;
    }
    return;
  }
}

void
risc::PathInstanceMapper::print_mapping()
{
  for(std::vector<PathWithID>::iterator
      iter  = path_to_id_mapping_.begin();
      iter != path_to_id_mapping_.end();
      iter++) {

    std::cout << (*iter).second << ", ";
    (*iter).first.print_instance_tree();
  }
}

int
risc::PathInstanceMapper::get_id(InstanceTree tree)
{
  int i = 0;
  for(std::vector<PathWithID>::iterator
      iter  = path_to_id_mapping_.begin();
      iter != path_to_id_mapping_.end();
      iter++, i++) {

    if((*iter).first == tree)
      return i;
  }

  assert("Not matching instance tree" && 0);
}

risc::InstanceTree
risc::PathInstanceMapper::get_instance_tree(
    int id,
    std::list<InstanceTree> potential_instances)
{
  for(std::list<InstanceTree>::iterator
      potential_iter  = potential_instances.begin();
      potential_iter != potential_instances.end();
      potential_iter++) {

    InstanceTree temp = (*potential_iter).get_parent();

    for(std::vector<PathWithID>::iterator
        mapping_iter  = path_to_id_mapping_.begin();
        mapping_iter != path_to_id_mapping_.end();
        mapping_iter++) {

      if((*mapping_iter).second == id &&
          (*mapping_iter).first == temp) {

        return *potential_iter;
      }
    }
  }

  assert(0 && "PathInstanceMapper::get_instance_tree: "
      && "Not event lookup possible ");
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
