#ifndef DATA_CONFLICT_TABLE_H_INCLUDED_
#define DATA_CONFLICT_TABLE_H_INCLUDED_

#include<vector>

#include "segment_graph.h"
#include "segment.h"
#include "conflict_table.h"

#include "../internal_representation/path_instance_mapper.h"
#include "data_conflict_ast_attributes.h"


namespace risc {

namespace sg {

/**
 * \class DataConflictTable
 * \brief This class represents the data conflict table among the individual
 *        segments. Two segments have a conflict if a least one variable is
 *        by the first segment and read by the second segment. A two
 *        dimensional array represents this array.
 */
class DataConflictTable: public ConflictTable {

public:

  /**
   * \brief    Default constructor for the DataConflictTable
   */
  DataConflictTable(
      SegmentGraph &graph,
      PathInstanceMapper &path_instance_mapper,
      bool debugging);

  /**
   * \function determine_conflict_table
   * \brief    This function determines conflicts among the segments
   */
  virtual void determine_conflict_table();

  /**
   * \function data_conflict_between
   * \brief    This function checks if there is a read write conflict between
   *           two segments.
   */
  bool data_conflict_between(VertexDescriptor first, VertexDescriptor second,
           int row, int column, int inst_id_first, int inst_id_second);

  /**
   * \function color_members_of_class
   * \brief    This function checks if the type of the given variable symbol is
   *           is a class or struct. If so, all the member variables which are
   *           not static will be set to the given color.
   */
  void color_members_of_class(SgVariableSymbol *var_symbol,
           DataConflictAstAttributes::Status color);

  /**
   * \function uncolor_members_of_class
   * \brief    This function checks if the type of the given variable symbol is
   *           is a class or struct. If so, all the member variables which are
   *           not static will be set to UNUSED.
   */
  void uncolor_members_of_class(SgVariableSymbol *var_symbol);

  /**
   * \function check_members_of_class
   * \brief    This function checks if the type of the given variable symbol is
   *           is a class or struct. If so, it will check if at least on member
   *           variable which is not static has the given color.
   */
  bool check_members_of_class(SgVariableSymbol *var_symbol,
           DataConflictAstAttributes::Status color);

  /**
   * \function get_mapped_variable
   * \brief    This function determines for a given variable the mapped
   *           variable. If variable is a reference, the mapped variable will
   *           be looked up. Otherwise, the original variable is returned.
   */
  SgVariableSymbol*
  get_mapped_variable(SgVariableSymbol *variable, int instance_id);
};


}; // end of namesapce sg

}; // end of namespace risc

#endif /* DATA_CONFLICT_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
