#include "segment.h"

int risc::sg::Segment::id_counter = 0;

risc::sg::Segment::Segment(SgNode const * const boundary_node):
  id_(id_counter),
  boundary_node_(boundary_node),
  in_conflict_with_all_other_segments(false),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched)
{
  Segment::id_counter++;
  debugging_ = true;
}

risc::sg::Segment::Segment(const Segment &other):
  id_(other.id_),
  boundary_node_(other.boundary_node_),
  expressions_(other.expressions_),
  in_conflict_with_all_other_segments(false),
  color_for_communication_graph_(Untouched),
  helper_for_graph_algorithms_(Untouched)
{
  debugging_ = true;
}

bool
risc::sg::Segment::operator==(const Segment &other)
{
  return id_ == other.id_;
}

void
risc::sg::Segment::add_expression(SgNode* node)
{
  expressions_.push_back(node);
}

void
risc::sg::Segment::merge(const Segment &other)
{
  if(*this == other)
    return;

  in_conflict_with_all_other_segments |=
    this->in_conflict_with_all_other_segments;

  for(Expressions::const_iterator
      iter  = other.expressions_.begin();
      iter != other.expressions_.end();
      iter++) {
    add_expression(*iter);
  }
}

void
risc::sg::Segment::analyze_segment()
{
  for(Segment::Expressions::iterator
      iter  = expressions_.begin();
      iter != expressions_.end();
      iter++) {

    if(isSgExpression(*iter)) {
      SgExpression *expr = isSgExpression(*iter);

      if(isSgAssignOp((*iter)->get_parent())) {
        SgAssignOp *assign_op = isSgAssignOp((*iter)->get_parent());
        if(assign_op->get_lhs_operand() == *iter) {

          analyze_expression(expr, true, false, false);
          continue;
        }
      }

      analyze_expression(expr, false, false, false);
    }

    if(isSgExprStatement(*iter)) {
      SgExprStatement *stmt_expr = isSgExprStatement(*iter);
      analyze_expression(stmt_expr->get_expression(), false, false, false);
    }

    if(isSgVariableDeclaration(*iter)) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      SgInitializedName *initialized_name = var_decl->get_decl_item(0);

      if(initialized_name->get_initptr()) {

        write_variables_.insert(
                             isSgVariableSymbol(
                                 initialized_name->
                                     search_for_symbol_from_symbol_table()));

        analyze_expression(initialized_name->get_initptr(),
            false, false, false);
      }
    }
  }
}


void
risc::sg::Segment::analyze_expression(SgNode *expression,
    bool left_of_assignment, bool read_write, bool member_access)
{
  #define ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(Operator)              \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_lhs_operand(), true, false, false);  \
    analyze_expression(op->get_lhs_operand(), false, false, false); \
    analyze_expression(op->get_rhs_operand(), false, false, false); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgAndAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgIorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgXorAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgLshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgRshiftAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgPlusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMinusAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgMultAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgDivAssignOp);
  ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP(SgModAssignOp);
  #undef ANALYZE_VARS_IN_COMPOUND_ASSIGN_OP

  #define ANALYZE_VARS_IN_BINARY_OP(Operator)                       \
  if(is##Operator(expression)) {                                    \
    Operator *op = is##Operator(expression);                        \
    analyze_expression(op->get_lhs_operand(), false, false, false); \
    analyze_expression(op->get_rhs_operand(), false, false, false); \
    return;                                                         \
  }
  ANALYZE_VARS_IN_BINARY_OP(SgLessThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterThanOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLessOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgGreaterOrEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgEqualityOp);
  ANALYZE_VARS_IN_BINARY_OP(SgNotEqualOp);
  ANALYZE_VARS_IN_BINARY_OP(SgLshiftOp);
  ANALYZE_VARS_IN_BINARY_OP(SgRshiftOp);

  ANALYZE_VARS_IN_BINARY_OP(SgBitAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgAndOp);
  ANALYZE_VARS_IN_BINARY_OP(SgOrOp);
  ANALYZE_VARS_IN_BINARY_OP(SgBitXorOp);

  ANALYZE_VARS_IN_BINARY_OP(SgAddOp);
  ANALYZE_VARS_IN_BINARY_OP(SgSubtractOp);
  ANALYZE_VARS_IN_BINARY_OP(SgDivideOp);
  ANALYZE_VARS_IN_BINARY_OP(SgMultiplyOp);
  ANALYZE_VARS_IN_BINARY_OP(SgModOp);

  ANALYZE_VARS_IN_BINARY_OP(SgCommaOpExp);
  #undef ANALYZE_VARS_IN_BINARY_OP


  #define ANALYZE_VARS_IN_UNARY_OP(Operator)                    \
  if(is##Operator(expression)) {                                \
    Operator *op = is##Operator(expression);                    \
    analyze_expression(op->get_operand(), false, false, false); \
    return;                                                     \
  }
  ANALYZE_VARS_IN_UNARY_OP(SgUnaryAddOp);
  ANALYZE_VARS_IN_UNARY_OP(SgMinusOp);
  ANALYZE_VARS_IN_UNARY_OP(SgAssignInitializer);
  ANALYZE_VARS_IN_UNARY_OP(SgNotOp);
  ANALYZE_VARS_IN_UNARY_OP(SgBitComplementOp);
  ANALYZE_VARS_IN_UNARY_OP(SgCastExp);
  #undef ANALYZE_VARS_IN_UNARY_OP

  #define ANALYZE_VARS_IN_UNARY_INC_DEC_OP(Operator)           \
  if(is##Operator(expression)) {                               \
    Operator *op = is##Operator(expression);                   \
    analyze_expression(op->get_operand(), false, true, false); \
    return;                                                    \
  }
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgPlusPlusOp);
  ANALYZE_VARS_IN_UNARY_INC_DEC_OP(SgMinusMinusOp);
  #undef ANALYZE_VARS_IN_UNARY_INC_DEC_OP

  #define ANALYZE_VARS_IN_ACCESS_OP(Operator)                     \
  if(is##Operator(expression)) {                                  \
    Operator *op = is##Operator(expression);                      \
    analyze_expression(op->get_rhs_operand(), left_of_assignment, \
                           read_write, member_access || false);   \
    analyze_expression(op->get_lhs_operand(), left_of_assignment, \
                           read_write,  true);                    \
    return;                                                       \
  }

  ANALYZE_VARS_IN_ACCESS_OP(SgArrowExp);
  ANALYZE_VARS_IN_ACCESS_OP(SgDotExp);

  ANALYZE_VARS_IN_ACCESS_OP(SgDotStarOp);
  ANALYZE_VARS_IN_ACCESS_OP(SgArrowStarOp);
  #undef ANALYZE_VARS_IN_ACCESS_OP

  if(isSgSizeOfOp(expression)) {
    SgSizeOfOp *size_of_op = isSgSizeOfOp(expression);
    analyze_expression(size_of_op->get_operand_expr(), false, false, false);
  }

  if(isSgVarRefExp(expression)) {
    SgVarRefExp *var_ref = isSgVarRefExp(expression);

    if(!member_access) {

      if(read_write) {
        write_variables_.insert(var_ref->get_symbol());
        read_variables_.insert(var_ref->get_symbol());
        return;
      }

      if(left_of_assignment) {
        write_variables_.insert(var_ref->get_symbol());
      } else {
        read_variables_.insert(var_ref->get_symbol());
      }
    }
  }

  if(isSgFunctionCallExp(expression)) {
    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expression);

    // Here we have to check if any parameter type is a reference or pointer
    // If so, the segment is with all other segments in conflict
    SgFunctionDeclaration *func_decl
      = func_call_exp->getAssociatedFunctionDeclaration();



    // we have to check because of virtual functions
    if(func_decl) {
      if(SageInterface::get_name(func_decl) != "notify" &&
          SageInterface::get_name(func_decl) != "fprintf" &&
          SageInterface::get_name(func_decl) != "exit" &&
          SageInterface::get_name(func_decl) != "sprintf") {

        SgFunctionParameterList *param_list = func_decl->get_parameterList();
        SgInitializedNamePtrList &init_name_list = param_list->get_args();

        for(SgInitializedNamePtrList::iterator
            iter  = init_name_list.begin();
            iter != init_name_list.end();
            iter++) {

          if(isSgPointerType((*iter)->get_type())) {

            if(debugging_) {

              conflicts_.insert(expression);
            }
            in_conflict_with_all_other_segments = true;
          }
        }
      }
    }

    SgExprListExp *arguments = func_call_exp->get_args();
    for(SgExpressionPtrList::iterator
        expr_iter  = arguments->get_expressions().begin();
        expr_iter != arguments->get_expressions().end();
        expr_iter++) {
      analyze_expression(*expr_iter, false, false, false);
    }
  }

  if(isSgPntrArrRefExp(expression)) {
    SgPntrArrRefExp *pointer_array_ref = isSgPntrArrRefExp(expression);

    analyze_expression(pointer_array_ref->get_lhs_operand(),
        left_of_assignment,
        read_write,
        member_access);
    analyze_expression(pointer_array_ref->get_rhs_operand(), false, false,
                           false);
  }

  if(isSgAssignOp(expression)) {
    SgAssignOp *assign_op = isSgAssignOp(expression);
    analyze_expression(assign_op->get_lhs_operand(), true, false, false);
    analyze_expression(assign_op->get_rhs_operand(), false, false, false);
  }

  if(isSgConditionalExp(expression)) {
    SgConditionalExp *conditional_exp = isSgConditionalExp(expression);
    analyze_expression(
        conditional_exp->get_conditional_exp(),
        left_of_assignment,
        read_write,
        false);
    analyze_expression(conditional_exp->get_true_exp(),
        left_of_assignment,
        read_write,
        false);
    analyze_expression(conditional_exp->get_false_exp(),
        left_of_assignment,
        read_write,
        false);
  }

  if(isSgPointerDerefExp(expression)) {
    SgPointerDerefExp *op = isSgPointerDerefExp(expression);
    analyze_expression(op->get_operand(), false, false, false);
    in_conflict_with_all_other_segments = true;
    if(debugging_) {

      conflicts_.insert(expression);
    }

    return;
  }

  if(isSgAddressOfOp(expression)) {
    SgAddressOfOp *op = isSgAddressOfOp(expression);
    analyze_expression(op->get_operand(), false, false, false);
    in_conflict_with_all_other_segments = true;
    if(debugging_) {

      conflicts_.insert(expression);
    }

    return;
  }

  assert(true && "Operator not supported");
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
